@@ -9,18 +9,41 @@
 package LaTeX::TOM::Parser;
 
 use strict;
+use base qw(
+    LaTeX::TOM::Node
+    LaTeX::TOM::Tree
+);
+use constant true  => 1;
+use constant false => 0;
 
-use base qw(LaTeX::TOM::Node);
-use base qw(LaTeX::TOM::Tree);
+use Carp qw(carp croak);
+use File::Basename qw(fileparse);
+
+our $VERSION = '0.07';
+
+my %error_handlers = (
+    0 => sub { warn "parse error: $_[0].\n" },
+    1 => sub { die  "parse error: $_[0].\n" },
+    2 => sub {},
+);
 
 # Constructor
 #
 sub new {
     my $class = shift;
 
-    my $self = {};
+    no strict 'refs';
 
-    bless $self, ref($class) || $class;
+    my $self = bless {
+        config => {
+            BRACELESS          => \%{"${class}::BRACELESS"},
+            INNERCMDS          => \%{"${class}::INNERCMDS"},
+            MATHENVS           => \%{"${class}::MATHENVS"},
+            MATHBRACKETS       => \%{"${class}::MATHBRACKETS"},
+            PARSE_ERRORS_FATAL =>  ${"${class}::PARSE_ERRORS_FATAL"},
+            TEXTENVS           => \%{"${class}::TEXTENVS"},
+        },
+    };
 
     $self->_init(@_);
 
@@ -30,17 +53,21 @@ sub new {
 # Set/reset "globals"
 #
 sub _init {
-    my $parser = $_[0];
+    my $parser = shift;
+    my ($parse_errors_fatal, $readinputs, $applymappings) = @_;
 
-    my $parse_errors_fatal = (defined $_[1] ? $_[1] : $parser->{config}{PARSE_ERRORS_FATAL});
-    my $readinputs = (defined $_[2] ? $_[2] : 0);
-    my $applymappings = (defined $_[3] ? $_[3] : 0);
+    my $retrieve_opt_default = sub
+    {
+        my ($opt, $default) = @_;
+        return $opt if defined $opt;
+        return $default;
+    };
 
     # set user options
     #
-    $parser->{readinputs} = $readinputs;
-    $parser->{applymappings} = $applymappings;
-    $parser->{PARSE_ERRORS_FATAL} = $parse_errors_fatal;
+    $parser->{readinputs}         = $retrieve_opt_default->($readinputs, 0);
+    $parser->{applymappings}      = $retrieve_opt_default->($applymappings, 0);
+    $parser->{PARSE_ERRORS_FATAL} = $retrieve_opt_default->($parse_errors_fatal, $parser->{config}{PARSE_ERRORS_FATAL});
 
     # init internal stuff
     #
@@ -68,17 +95,13 @@ sub parseFile {
     $parser->{file} = $filename;        # file name member data
     my $tree = {};                      # init output tree
 
-    # read in text from file
+    # read in text from file or bomb out
     #
-    my $text = _readFile($filename);
+    my $text = _readFile($filename, true);
 
-    # do the parse or bomb out
+    # do the parse
     #
-    if ($text) {
-        $tree = $parser->parse($text);
-    } else {
-        die "Could not read file $filename !";
-    }
+    $tree = $parser->parse($text);
 
     return $tree;
 }
@@ -87,36 +110,39 @@ sub parseFile {
 #
 sub parse {
     my $parser = shift;
-
-    my $text = shift;
+    my ($text) = @_;
 
     # first half of parsing (goes up to finding commands, reading inputs)
     #
     my ($tree, $bracehash) = $parser->_parseA($text);
-
-    #print "done with _parseA\n";
-    #$tree->print();
+    _debug(
+        'done with _parseA',
+         sub { $tree->_warn() },
+    );
 
     # handle mappings
     #
     $parser->_applyMappings($tree) if $parser->{applymappings};
-
-    #print "done with _applyMappings\n";
-    #$tree->print();
+    _debug(
+        'done with _applyMappings',
+         sub { $tree->_warn() },
+    );
 
     # second half of parsing (environments)
     #
     $parser->_parseB($tree);
-
-    #print "done with _parseB\n";
-    #$tree->print();
+    _debug(
+        'done with _parseB',
+         sub { $tree->_warn() },
+    );
 
     # once all the above is done we can propegate math/plaintext modes down
     #
     $parser->_propegateModes($tree, 0, 0);   # math = 0, plaintext = 0
-
-    #print "done with _propegateModes\n";
-    #$tree->print();
+    _debug(
+        'done with _propegateModes',
+         sub { $tree->_warn() },
+    );
 
     # handle kooky \[ \] math mode
     #
@@ -127,8 +153,10 @@ sub parse {
         $parser->{MATHBRACKETS}->{'\\['} = '\\]';  # put back in brackets list for
         $parser->{MATHBRACKETS}->{'\\('} = '\\)';  # printing purposes.
     }
-
-    #$tree->print();
+    _debug(
+        undef,
+        sub { $tree->_warn() },
+    );
 
     $tree->listify;     # add linked-list stuff
 
@@ -162,7 +190,7 @@ sub _stage1 {
 
     my @nodes = _getTextAndCommentNodes($text, 0, length($text));
 
-    return LaTeX::TOM::Tree->new([@nodes], $parser);
+    return LaTeX::TOM::Tree->new([@nodes]);
 }
 
 # this stage parses the braces ({}) and adds the corresponding structure to
@@ -196,7 +224,8 @@ sub _stage2 {
 
         if ($node->{type} eq 'TEXT') {
 
-         #warn "parseStage2: looking at text node: [$node->{content}]";
+         _debug("parseStage2: looking at text node: [$node->{content}]", undef);
+
          my ($nextpos, $brace) = _findbrace($node->{content}, $pos);
          while ($nextpos != -1) {
 
@@ -204,7 +233,7 @@ sub _stage2 {
 
             # handle left brace
             if ($brace eq '{') {
-                #warn "found '{' at position $nextpos, leftcount is $leftcount";
+                _debug("found '{' at position $nextpos, leftcount is $leftcount", undef);
                 if ($leftcount == 0) {
                     $leftpos = $nextpos;
                     $leftidx = $i
@@ -215,7 +244,7 @@ sub _stage2 {
             # handle right brance
             elsif ($brace eq '}') {
 
-                #warn "found '}' at position $nextpos , leftcount is $leftcount";
+                _debug("found '}' at position $nextpos, leftcount is $leftcount", undef);
                 my $rightpos = $nextpos;
                 $leftcount--;
 
@@ -234,7 +263,7 @@ sub _stage2 {
                             {type => 'GROUP',
                              start => $textnode2->{start} - 1,
                              end => $textnode2->{end} + 1,
-                             children => LaTeX::TOM::Tree->new([$textnode2], $parser)
+                             children => LaTeX::TOM::Tree->new([$textnode2]),
                             });
 
                         # splice the new subtree into the old location
@@ -271,10 +300,10 @@ sub _stage2 {
                              start => $textnode2->{start} - 1,
                              end => $textnode3->{end} + 1,
                              children => LaTeX::TOM::Tree->new(
-                                [$textnode2, 
+                                [$textnode2,
                                  @removed,
                                  $textnode3]),
-                                 $parser});
+                            });
 
                         # replace the two original text nodes with the leftover left and
                         # right portions, as well as the group node with everything in
@@ -301,13 +330,8 @@ sub _stage2 {
                 # check for '}'-based error
                 #
                 if ($leftcount < 0) {
-                    if ($parser->{PARSE_ERRORS_FATAL} == 1) {
-                        die "parse error: '}' before '{' at ".($spos+$rightpos).".";
-                    } elsif ($parser->{PARSE_ERRORS_FATAL} == 0) {
-                        warn "parse error: '}' before '{' at ".($spos+$rightpos).".";
-                    }
-                    # reset and continue
-                    $leftcount = 0;
+                    $error_handlers{$parser->{PARSE_ERRORS_FATAL}}->("'}' before '{' at " . ($spos + $rightpos));
+                    $leftcount = 0; # reset and continue
                 }
             } # right brace
 
@@ -322,12 +346,9 @@ sub _stage2 {
     # check for extra '{' parse error
     #
     if ($leftcount > 0) {
-        my $spos = $tree->{nodes}[$leftidx]->{start};	# get text start position 
-        if ($parser->{PARSE_ERRORS_FATAL} == 1) {
-            die "parse error: unmatched '{' at ".($spos+$leftpos).".";
-        } elsif ($parser->{PARSE_ERRORS_FATAL} == 0) {
-            warn "parse error: unmatched '{' at ".($spos+$leftpos).".";
-        }
+        my $spos = $tree->{nodes}[$leftidx]->{start}; # get text start position
+        $error_handlers{$parser->{PARSE_ERRORS_FATAL}}->("unmatched '{' at " . ($spos + $leftpos));
+
         # try to continue on, after the offending brace
         $parser->_stage2($tree, $bracehash, $leftidx, $leftpos + 1);
     }
@@ -357,7 +378,7 @@ sub _stage3 {
             # due to the previous parsing stages. if found, the parent node is
             # promoted to a command.
             #
-            if ($text =~ /^\s*\\(\w+\*?)/ && defined $parent && defined $parser->{config}{INNERCMDS}->{$1}) {
+            if ($text =~ /^\s*\\(\w+\*?)/ && defined $parent && $parser->{config}{INNERCMDS}->{$1}) {
                 my $command = $1;
 
                 # if the parent is already a command node, we have to make a new
@@ -375,7 +396,7 @@ sub _stage3 {
                          children => $parent->{children} });
 
                     # point parent to it
-                    $parent->{children} = LaTeX::TOM::Tree->new([$newnode], $parser);
+                    $parent->{children} = LaTeX::TOM::Tree->new([$newnode]);
 
                     # start over at this level (get additional inner commands)
                     $parent = $newnode;
@@ -410,7 +431,7 @@ sub _stage3 {
 
                 my $tag = $1;
 
-                #print "found text node [$text] with command tag [$tag]\n";
+                _debug("found text node [$text] with command tag [$tag]", undef);
 
                 # remove the text
                 $node->{content} =~ s/\\\w+\*?\s*(?:\[.*?\])?\s*$//os;
@@ -439,7 +460,7 @@ sub _stage3 {
                 my $command = $2;
                 my $param = $3;
 
-                if (defined $parser->{config}{BRACELESS}->{$command}) {
+                if ($parser->{config}{BRACELESS}->{$command}) {
                  # warn "found braceless command $command with param $param";
 
                     # get location to split from node text
@@ -466,7 +487,8 @@ sub _stage3 {
                          command => $command,
                          start => $node->{start} + $a,
                          end => $node->{start} + $b,
-                         children => LaTeX::TOM::Tree->new([$newchild]), $parser });
+                         children => LaTeX::TOM::Tree->new([$newchild]),
+                        });
 
                     $parser->{USED_COMMANDS}->{$commandnode->{command}} = 1;
 
@@ -517,20 +539,20 @@ sub _stage4 {
         # see if this is a "\begin" command node
         if ($node->{type} eq 'COMMAND' && $node->{command} eq 'begin') {
 
-            #warn "parseStage4: found a begin COMMAND node, ".$node->{children}->{nodes}[0]->{content}." @ $node->{start}";
+            _debug("parseStage4: found a begin COMMAND node, $node->{children}->{nodes}[0]->{content} @ $node->{start}", undef);
 
             # start a new "stack"
             if ($bcount == 0) {
                 $bidx = $i;
                 $bcount++;
                 $class = $node->{children}->{nodes}->[0]->{content}; 
-                #warn "parseStage4: opening environment tag found, class = $class";
+                _debug("parseStage4: opening environment tag found, class = $class", undef);
             }
 
             # add to the "stack"
             elsif ($node->{children}->{nodes}->[0]->{content} eq $class) {
                 $bcount++;
-                #warn "parseStage4: incrementing tag count for $class";
+                _debug("parseStage4: incrementing tag count for $class", undef);
             }
         }
 
@@ -540,14 +562,14 @@ sub _stage4 {
                $node->{children}->{nodes}->[0]->{content} eq $class) {
 
             $bcount--;
-            #warn "parseStage4: decrementing tag count for $class";
+            _debug("parseStage4: decrementing tag count for $class", undef);
 
             # we found our closing "\end" tag. replace everything with the proper
             # ENVIRONMENT tag and subtree.
             #
             if ($bcount == 0) {
 
-                #warn "parseStage4: closing environment $class";
+                _debug("parseStage4: closing environment $class", undef);
 
                 # first we must take everything between the "\begin" and "\end" 
                 # nodes and put them in a new array, removing them from the old one
@@ -563,9 +585,10 @@ sub _stage4 {
                      end => $end,
                      ostart => $start - length('begin') - length($class) - 2,
                      oend => $end + length('end') + length($class) + 2,
-                     children => LaTeX::TOM::Tree->new([@newarray]), $parser });
+                     children => LaTeX::TOM::Tree->new([@newarray]),
+                    });
 
-                if (defined $parser->{config}{MATHENVS}->{$envnode->{class}}) {
+                if ($parser->{config}{MATHENVS}->{$envnode->{class}}) {
                     $envnode->{math} = 1;
                 }
 
@@ -591,11 +614,9 @@ sub _stage4 {
 
     # parse error if we're missing an "\end" tag.
     if ($bcount > 0) {
-        if ($parser->{PARSE_ERRORS_FATAL} == 1) {
-            die "parse error: missing \\end{$class} for \\begin{$class} at position ".$tree->{nodes}[$bidx]->{end}." !";
-        } elsif ($parser->{PARSE_ERRORS_FATAL} == 0) {
-            warn "parse error: missing \\end{$class} for \\begin{$class} at position ".$tree->{nodes}[$bidx]->{end}." !";
-        }
+        $error_handlers{$parser->{PARSE_ERRORS_FATAL}}->(
+            "missing \\end{$class} for \\begin{$class} at position $tree->{nodes}[$bidx]->{end}"
+        );
     }
 }
 
@@ -650,7 +671,7 @@ sub _stage5_r {
                     $leftpos = _findsymbol($node->{content}, $left, $pos);
 
                     if ($leftpos != -1) {
-                        #print "found (left) $left in [$node->{content}]\n";
+                        _debug("found (left) $left in [$node->{content}]", undef);
                         $leftidx = $i;
                         $pos = $leftpos + 1; # next pos to search from
                     }
@@ -665,7 +686,7 @@ sub _stage5_r {
 
                         # we have to split the text node into 3 parts
                         if ($leftidx == $i) {
-                            #print "splitwithin: found (right) $right in [$node->{content}]\n";
+                            _debug("splitwithin: found (right) $right in [$node->{content}]", undef);
 
                             my ($leftnode, $textnode3) = $node->split($rightpos, $rightpos + length($right) - 1);
                             my ($textnode1, $textnode2) = $leftnode->split($leftpos, $leftpos + length($left) - 1);
@@ -681,7 +702,8 @@ sub _stage5_r {
                                 ostart => $startpos + $leftpos - length($left) + 1,
                                 end => $startpos + $rightpos,
                                 oend => $startpos + $rightpos + length($right) - 1,
-                                children => LaTeX::TOM::Tree->new([$textnode2]), $parser });
+                                children => LaTeX::TOM::Tree->new([$textnode2]),
+                                });
 
                             splice @{$tree->{nodes}}, $i, 1, $textnode1, $mathnode, $textnode3;
 
@@ -691,7 +713,7 @@ sub _stage5_r {
                         # split across nodes
                         else {
 
-                            #print "splitacross: found (right) $right in [$node->{content}]\n";
+                            _debug("splitacross: found (right) $right in [$node->{content}]", undef);
 
                             # create new set of 4 smaller text nodes from the original two
                             # that contain the left and right delimeters
@@ -719,7 +741,8 @@ sub _stage5_r {
                                 children => LaTeX::TOM::Tree->new(
                                 [$textnode2,
                                  @remnodes,
-                                 $textnode3]), $parser });
+                                 $textnode3]),
+                                });
 
                             # replace (TEXT_A, ... , TEXT_B) with the mathnode created above
                             splice @{$tree->{nodes}}, $leftidx, 2, $textnode1, $mathnode, $textnode4;
@@ -738,12 +761,8 @@ sub _stage5_r {
                     my $rightpos = _findsymbol($node->{content}, $right, $pos);
 
                     if ($rightpos != -1) {
-                        my $startpos = $node->{start}; # get text start position 
-                        if ($parser->{PARSE_ERRORS_FATAL} == 1) {
-                            die "parse error: unmatched '$right' at ".($startpos+$rightpos).".";
-                        } elsif ($parser->{PARSE_ERRORS_FATAL} == 0) {
-                            warn "parse error: unmatched '$right' at ".($startpos+$rightpos).".";
-                        }
+                        my $startpos = $node->{start}; # get text start position
+                        $error_handlers{$parser->{PARSE_ERRORS_FATAL}}->("unmatched '$right' at " . ($startpos + $rightpos));
                     }
                 }
             } # if TEXT
@@ -755,10 +774,12 @@ sub _stage5_r {
                              ($node->{type} eq 'COMMAND' && $node->{command} =~ /^verb/) ||
                              ($node->{type} eq 'ENVIRONMENT' && $node->{class} =~ /^verbatim/))) {
 
-                #print "Recurring into $node->{type} node ";
-                #print "$node->{command}" if ($node->{type} eq 'COMMAND');
-                #print "$node->{class}" if ($node->{type} eq 'ENVIRONMENT');
-                #print "\n";
+                if ($LaTeX::TOM::DEBUG) {
+                    my $message  = "Recurring into $node->{type} node ";
+                       $message .= $node->{command} if ($node->{type} eq 'COMMAND');
+                       $message .= $node->{class}   if ($node->{type} eq 'ENVIRONMENT');
+                    _debug($message, undef);
+                }
 
                 $parser->_stage5_r($node->{children}, $left, $right, $caremath);
             }
@@ -767,11 +788,7 @@ sub _stage5_r {
 
         if ($leftpos != -1) {
             my $startpos = $tree->{nodes}[$leftidx]->{start};   # get text start position
-            if ($parser->{PARSE_ERRORS_FATAL} == 1) {
-                die "parse error: unmatched '$left' at ".($startpos+$leftpos).".";
-            } elsif ($parser->{PARSE_ERRORS_FATAL} == 0) {
-                warn "parse error: unmatched '$left' at ".($startpos+$leftpos).".";
-            }
+            $error_handlers{$parser->{PARSE_ERRORS_FATAL}}->("unmatched '$left' at " . ($startpos + $leftpos));
         }
 }
 
@@ -807,18 +824,20 @@ sub _propegateModes {
             # handle math or plain text forcing envs
             #
             if ($node->{type} eq 'ENVIRONMENT' || $node->{type} eq 'COMMAND') {
-                if (defined $parser->{config}{MATHENVS}->{$node->{class}} || 
-                    defined $parser->{config}{MATHENVS}->{"$node->{class}*"}) {
-
+                if (defined $node->{class} && (
+                    $parser->{config}{MATHENVS}->{$node->{class}} ||
+                    $parser->{config}{MATHENVS}->{"$node->{class}*"})
+                   )
+                {
                     $mathflag = 1;
                     $plaintextflag = 0;
                 }
                 elsif (($node->{type} eq 'COMMAND' && 
-                                (defined $parser->{config}{TEXTENVS}->{$node->{command}} ||
-                                 defined $parser->{config}{TEXTENVS}->{"$node->{command}*"})) ||
+                                ($parser->{config}{TEXTENVS}->{$node->{command}} ||
+                                 $parser->{config}{TEXTENVS}->{"$node->{command}*"})) ||
                              ($node->{type} eq 'ENVIRONMENT' && 
-                                (defined $parser->{config}{TEXTENVS}->{$node->{class}} ||
-                                 defined $parser->{config}{TEXTENVS}{"$node->{command}*"}))
+                                ($parser->{config}{TEXTENVS}->{$node->{class}} ||
+                                 $parser->{config}{TEXTENVS}{"$node->{command}*"}))
                             ) {
 
                     $mathflag = 0;
@@ -873,10 +892,10 @@ sub _applyMapping {
 
         # begin environment nodes
         #
-        if ($node->{type} eq 'COMMAND' &&
-                $node->{command} eq 'begin' &&
-                $node->{children}->{nodes}[0]->{content} eq $mapping->{name}) {
-
+        if ($node->{type}                            eq 'COMMAND'
+         && $node->{command}                         eq 'begin'
+         && $node->{children}->{nodes}[0]->{content} eq $mapping->{name}
+        ) {
             # grab the nparams next group nodes as parameters
             #
             my @params = ();
@@ -903,36 +922,36 @@ sub _applyMapping {
             my $applied = _applyParamsToTemplate($mapping->{btemplate}, @params);
 
             # splice in the result
-            splice @{$tree->{nodes}}, $i, $j, @$applied;
+            splice @{$tree->{nodes}}, $i, $j, @{$applied->{nodes}};
 
             # skip past all the new stuff
-            $i += scalar @$applied - 1;
+            $i += scalar @{$applied->{nodes}} - 1;
         }
 
         # end environment nodes
         #
-        elsif ($node->{type} eq 'COMMAND' &&
-                     $node->{command} eq 'end' &&
-                     $node->{children}->{nodes}[0]->{content} eq $mapping->{name}) {
-
+        elsif ($node->{type}                            eq 'COMMAND'
+            && $node->{command}                         eq 'end'
+            && $node->{children}->{nodes}[0]->{content} eq $mapping->{name}
+        ) {
             # make new subtree (no params)
             my $applied = $mapping->{etemplate}->copy();
 
             # splice in the result
-            splice @{$tree->{nodes}}, $i, 1, @$applied;
+            splice @{$tree->{nodes}}, $i, 1, @{$applied->{nodes}};
 
             # skip past all the new stuff
-            $i += scalar @$applied - 1;
+            $i += scalar @{$applied->{nodes}} - 1;
 
             $applications++; # only count end environment nodes
         }
 
         # newcommand nodes
         #
-        elsif ($node->{type} eq 'COMMAND' &&
-                $node->{command} eq $mapping->{name} &&
-                $mapping->{nparams}) {
-
+        elsif ($node->{type}       eq 'COMMAND'
+            && $node->{command}    eq $mapping->{name}
+            && $mapping->{nparams}
+        ) {
             my @params = ();
 
             # children of COMMAND node will be first parameter
@@ -961,10 +980,10 @@ sub _applyMapping {
             my $applied = _applyParamsToTemplate($mapping->{template}, @params);
 
             # splice in the result
-            splice @{$tree->{nodes}}, $i, $j, @$applied;
+            splice @{$tree->{nodes}}, $i, $j, @{$applied->{nodes}};
 
             # skip past all the new stuff
-            $i += scalar @$applied - 1;
+            $i += scalar @{$applied->{nodes}} - 1;
 
             $applications++;
         }
@@ -981,21 +1000,21 @@ sub _applyMapping {
              while (($wordend && $text =~ /\\\Q$command\E(\W|$)/g) ||
                             (!$wordend && $text =~ /\\\Q$command\E/g)) {
 
-                 #warn "found occurrence of mapping $command";
+                 _debug("found occurrence of mapping $command", undef);
 
                  my $idx = index $node->{content}, '\\' . $command, 0;
 
                  # split the text node at that command
-                 my ($leftnode, $rightnode) = splitTextNode($node, $idx, $idx + length($command));
+                 my ($leftnode, $rightnode) = $node->split($idx, $idx + length($command));
 
                  # copy the mapping template
                  my $applied = $mapping->{template}->copy();
 
                  # splice the new nodes in
-                 splice @{$tree->{nodes}}, $i, 1, $leftnode, @$applied, $rightnode;
+                 splice @{$tree->{nodes}}, $i, 1, $leftnode, @{$applied->{nodes}}, $rightnode;
 
                  # adjust i so we end up on rightnode when we're done
-                 $i += scalar @$applied + 1;
+                 $i += scalar @{$applied->{nodes}} + 1;
 
                  # get the next node
                  $node = $tree->{$node}[$i];
@@ -1037,9 +1056,9 @@ sub _applyMappings {
             next if (!$mapping->{name}); # skip fragged commands
 
             if ($parser->{USED_COMMANDS}->{$mapping->{name}}) {
-                #print "applying (nc) mapping $mapping->{name}\n";
+                _debug("applying (nc) mapping $mapping->{name}", undef);
             } else {
-                #print "NOT applying (nc) mapping $mapping->{name}\n";
+                _debug("NOT applying (nc) mapping $mapping->{name}", undef);
                 next;
             }
 
@@ -1047,7 +1066,7 @@ sub _applyMappings {
             #
             $parser->{MAPPEDCMDS}->{"\\$mapping->{name}"} = 1;
 
-            #print "found a mapping with name $mapping->{name}, $mapping->{nparams} params\n";
+            _debug("found a mapping with name $mapping->{name}, $mapping->{nparams} params", undef);
 
             # remove the mapping declaration
             #
@@ -1056,10 +1075,9 @@ sub _applyMappings {
             # apply the mapping
             my $count = $parser->_applyMapping($tree, $mapping, $i);
 
-            #if ($count > 0) {
-            #   print "printing altered subtree\n";
-            #   $tree->print();
-            #}
+            if ($count > 0) {
+                _debug("printing altered subtree", sub { $tree->_warn() });
+            }
 
             $i--; # since we removed the cmd node, check this index again
         }
@@ -1073,7 +1091,7 @@ sub _applyMappings {
             my $mapping = $parser->_makeEnvMapping($tree, $i);
             next if (!$mapping->{name}); # skip fragged commands.
 
-            #print "applying (ne) mapping $mapping->{name}\n";
+            _debug("applying (ne) mapping $mapping->{name}", undef);
 
             # remove the mapping declaration
             #
@@ -1090,7 +1108,7 @@ sub _applyMappings {
                  $prev->{type} eq 'TEXT' &&
                  $prev->{content} =~ /\\def\s*$/o) {
 
-             #print "found def style mapping $node->{command}\n";
+             _debug("found def style mapping $node->{command}", undef);
 
              # remove the \def
              $prev->{content} =~ s/\\def\s*$//o;
@@ -1098,15 +1116,15 @@ sub _applyMappings {
              # make the mapping
              my $mapping = {name => $node->{command},
                 nparams => 0,
-                template => copyTree($node->{children}),
+                template => $node->{children}->copy(),
                 type => 'command'};
 
              next if (!$mapping->{name}); # skip fragged commands
 
              if ($parser->{USED_COMMANDS}->{$mapping->{name}}) {
-                 #print "applying (def) mapping $mapping->{name}\n";
+                 _debug("applying (def) mapping $mapping->{name}", undef);
              } else {
-                 #print "NOT applying (def) mapping $mapping->{name}\n";
+                 _debug("NOT applying (def) mapping $mapping->{name}", undef);
                  next;
              }
 
@@ -1114,8 +1132,7 @@ sub _applyMappings {
              #
              $parser->{MAPPEDCMDS}->{"\\$mapping->{name}"} = 1;
 
-             #print "template is \n";
-             #$mapping->{template}->print();
+             _debug("template is", sub { $mapping->{template}->_warn() });
 
              # remove the command node
              splice @{$tree->{nodes}}, $i, 1;
@@ -1147,59 +1164,64 @@ sub _addInputs {
 
         my $node = $tree->{nodes}[$i];
 
-        if ($node->{type} eq 'COMMAND' &&
-            $node->{command} eq 'input') {
-
+        if ($node->{type}    eq 'COMMAND'
+         && $node->{command} eq 'input'
+        ) {
             my $file = $node->{children}->{nodes}[0]->{content};
-            next if ($file =~ /pstex/);	# ignore pstex images
+            next if $file =~ /pstex/; # ignore pstex images
 
-            #print "reading input file $file\n";
+            _debug("reading input file $file", undef);
+
+            my $contents;
+            my $filename = fileparse($file);
+            my $has_extension = qr/\.\S+$/;
 
             # read in contents of file
-            my $contents = _readFile($file);
-            if ($contents eq '' && not $file =~ /\.tex$/) {
+            if (-e $file && $filename =~ $has_extension) {
+                $contents = _readFile($file);
+            }
+            elsif ($filename !~ $has_extension) {
                 $file = "$file.tex";
-                $contents = _readFile($file); 
+                $contents = _readFile($file) if -e $file;
             }
 
             # dump Psfig/TeX files, they aren't useful to us and have
             # nonconforming syntax. Use declaration line as our heuristic.
             #
-            if ($contents =~ /^\%\s*Psfig\/TeX\s*$/m) {
-                $contents = '';
-                warn "ignoring Psfig input '$file'";
+            if (defined $contents
+                     && $contents =~ m!^ \% \s*? Psfig/TeX \s* $!mx
+            ) {
+                undef $contents;
+                carp "ignoring Psfig input `$file'";
             }
 
             # actually do the parse of the sub-content
             #
-            if ($contents) {
+            if (defined $contents) {
                 # parse into a tree
                 my ($subtree,) = $parser->_basicparse($contents, $parser->{PARSE_ERRORS_FATAL});
 
                 # replace \input command node with subtree
-                splice @{$tree->{nodes}}, $i, 1, @$subtree;
+                splice @{$tree->{nodes}}, $i, 1, @{$subtree->{nodes}};
 
                 # step back
                 $i--;
             }
         }
-        elsif ($node->{type} eq 'COMMAND' &&
-             $node->{command} eq 'bibliography') {
-
+        elsif ($node->{type}    eq 'COMMAND'
+            && $node->{command} eq 'bibliography'
+        ) {
              # try to find a .bbl file
              #
-             my @FILES = <*>;
-             foreach my $file (@FILES) {
-                 if ($file =~ /\.bbl$/) {
+             foreach my $file (<*.bbl>) {
 
-                     my $contents = _readFile($file);
+                 my $contents = _readFile($file);
 
-                     if ($contents) {
+                 if (defined $contents) {
 
-                         my ($subtree,) = $parser->_basicparse($contents, $parser->{PARSE_ERRORS_FATAL});
-                         splice @{$tree->{nodes}}, $i, 1, @$subtree;
-                         $i--;
-                     }
+                     my ($subtree,) = $parser->_basicparse($contents, $parser->{PARSE_ERRORS_FATAL});
+                     splice @{$tree->{nodes}}, $i, 1, @{$subtree->{nodes}};
+                     $i--;
                  }
              }
         }
@@ -1235,11 +1257,11 @@ sub _parseB {
 
     $parser->_stage4($tree);
 
-    #print "done with parseStage4\n";
+    _debug("done with parseStage4", undef);
 
     $parser->_stage5($tree, 0);
 
-    #print "done with parseStage5\n";
+    _debug("done with parseStage5", undef);
 }
 
 ###############################################################################
@@ -1351,7 +1373,7 @@ sub _findbrace {
         $realbrace = 1;
         $index_c = index $text, '}', $pos_c;
 
-        if (substr($text, $index_c - 1, 1) eq ' ') {
+        if (($index_c - 1) >= 0 && substr($text, $index_c - 1, 1) eq ' ') {
             $pos_c = $index_c + 1;
             $index_c = -1;
         }
@@ -1428,7 +1450,7 @@ sub _duplicateParam {
         $nodecopy->{children} = $subtree; # set the child pointer to new subtree
 
         # return a new tree with the new node (subtree) as its only element
-        return LaTeX::TOM::Tree->new([$nodecopy], $parser);
+        return LaTeX::TOM::Tree->new([$nodecopy]);
     }
 
     return undef;
@@ -1581,9 +1603,9 @@ sub _applyParamsToTemplate_r {
     my $template = shift;
     my @params = @_;
 
-    for (my $i = 0; $i < @$template; $i++) {
+    for (my $i = 0; $i < @{$template->{nodes}}; $i++) {
 
-        my $node = $template->[$i];
+        my $node = $template->{nodes}[$i];
 
         if ($node->{type} eq 'TEXT') {
 
@@ -1611,11 +1633,11 @@ sub _applyParamsToTemplate_r {
                 # splice the new text nodes, along with the parameter subtree, into
                 # the old location
                 #
-                splice @$template, $i, 1, $leftnode, @$param, $rightnode;
+                splice @{$template->{nodes}}, $i, 1, $leftnode, @{$param->{nodes}}, $rightnode;
 
                 # skip forward to where $rightnode is in $template on next iteration
                 #
-                $i += scalar @$param;
+                $i += scalar @{$param->{nodes}};
             }
         }
 
@@ -1633,76 +1655,71 @@ sub _applyParamsToTemplate_r {
 # '%' prefixed LaTeX comment lines
 #
 sub _getTextAndCommentNodes {
-    my $text = shift;
-    my $begins = shift;
-    my $ends = shift;
+    my ($text, $begins, $ends) = @_;
 
-    my $nodetext = substr $text, $begins, $ends - $begins;
+    my $node_text = substr $text, $begins, $ends - $begins;
 
-    #warn "getTextAndCommentNodes: looking at [$nodetext]";
+    _debug("getTextAndCommentNodes: looking at [$node_text]", undef);
 
-    my @lines = split (/((?: *(?<!\\)\%[^\n]*\n)+)/m, $nodetext);
+    my $make_node = sub {
+        my ($mode_type, $begins, $start_pos, $output) = @_;
 
-    my @nodes = ();
+        return LaTeX::TOM::Node->new({
+            type    => uc $mode_type,
+            start   => $begins + $start_pos,
+            end     => $begins + $start_pos + length($output) - 1,
+            content => $output,
+        });
+    };
 
-    my $startpos = 0;
-    my @out = ();
-    my $comment = 0; # comment mode
-    my $first = 1;
+    my @lines = split (/(
+       (?:\s*     # whitespace
+         (?<!\\)  # unescaped
+         \%[^\n]* # comment
+       \n)+       # newline
+    )/mx, $node_text);
+
+    my @nodes;
+
+    my $start_pos = 0;
+    my $output;
+    my $mode_type;
+    my $first = true;
 
     foreach my $line (@lines) {
 
-        my $mode;
+         my $line_type = (
+                 $line =~ /^\s*\%/
+         && $node_text !~ /
+                           \\begin\{verbatim\}
+                             .* \Q$line\E .*
+                           \\end\{verbatim\}
+                          /sx
+        ) ? 'comment' : 'text';
 
-        # set current mode based on whether this is a comment
-        if ($line =~ /^\s*\%/ && $nodetext !~ /\\begin\{verbatim\}.*\Q$line\E.*\\end\{verbatim\}/s) {
-            $mode = 1;  # comment
-        } else {
-            $mode = 0;  # not-comment (regular text)
-        }
+        # if type stays the same, add to output and do nothing
+        if ($first || $line_type eq $mode_type) {
 
-        # if mode stays the same, add to output and do nothing
-        if ($mode == $comment || $first) { 
-
-            push @out, $line;
+            $output .= $line;
 
             # handle turning off initialization stuff
-            if ($first) {
-                $first = 0;
-                $comment = $mode;
-            }
+            $first &&= false;
+            $mode_type ||= $line_type;
         }
 
-        # if mode changes, make new node from current chunk, change mode
-        # flag, and start a new chunk
+        # if type changes, make new node from current chunk, change mode type
+        # and start a new chunk
         else {
-        my $output = join ('', @out);
+            push @nodes, $make_node->($mode_type, $begins, $start_pos, $output);
 
-            my $newnode = LaTeX::TOM::Node->new( 
-                {type => $comment ? 'COMMENT' : 'TEXT',
-                 start => $begins + $startpos,
-                 end => $begins + $startpos + length($output) - 1,
-                 content => $output});
+            $start_pos += length($output); # update start position
+            $output = $line;
 
-            push @nodes, $newnode;
-
-            $startpos += length($output); # update start position 
-            @out = ($line);
-
-            $comment = $mode;
+            $mode_type = $line_type;
         }
     }
 
-    # make node from last chunk 
-    my $output = join('', @out);
-
-    my $newnode = LaTeX::TOM::Node->new(
-        {type => $comment ? 'COMMENT' : 'TEXT',
-         start => $begins + $startpos,
-         end => $begins + $startpos + length($output) - 1,
-         content => $output});
-
-    push @nodes, $newnode;
+    push @nodes, $make_node->($mode_type, $begins, $start_pos, $output) if defined $output;
 
     return @nodes;
 }
@@ -1710,19 +1727,35 @@ sub _getTextAndCommentNodes {
 # Read in the contents of a text file on disk. Return in string scalar.
 #
 sub _readFile {
-    my $file = shift || '/dev/stdin';
+    my ($file, $raise_error) = @_;
 
-    my $contents = "";
+    $raise_error ||= false;
 
-    open INFILE, $file or return "";
+    my $opened = open(my $fh, '<', $file);
 
-    while (my $line = <INFILE>) {
-        $contents .= $line;
+    unless ($opened) {
+        croak "Cannot open $file: $!" if $raise_error;
+        return undef;
     }
 
-    close INFILE;
+    my $contents = do { local $/; <$fh> };
+    close($fh);
 
     return $contents;
 }
 
+sub _debug {
+    my ($message, $code) = @_;
+
+    my $DEBUG = $LaTeX::TOM::DEBUG;
+
+    return unless $DEBUG >= 1 && $DEBUG <= 2;
+
+    my ($filename, $line) = (caller)[1,2];
+    my $caller = join ':', (fileparse($filename))[0], $line;
+
+    warn "$caller: $message\n" if $DEBUG >= 1 && defined $message;
+    $code->()                  if $DEBUG == 2 && defined $code;
+}
+
 1;
